"use strict";
var $$state$$ = require("./state"), $$tokentype$$ = require("./tokentype"), $$whitespace$$ = require("./whitespace");

class TokContext {
  constructor(token, isExpr, preserveSpace, override) {
    this.token = token
    this.isExpr = !!isExpr
    this.preserveSpace = !!preserveSpace
    this.override = override
  }
}

const types = {
  b_stat: new TokContext("{", false),
  b_expr: new TokContext("{", true),
  b_tmpl: new TokContext("${", true),
  p_stat: new TokContext("(", false),
  p_expr: new TokContext("(", true),
  q_tmpl: new TokContext("`", true, true, p => p.readTmplToken()),
  f_expr: new TokContext("function", true)
};

const pp = $$state$$.Parser.prototype

pp.initialContext = function() {
  return [types.b_stat]
}

pp.braceIsBlock = function(prevType) {
  if (prevType === $$tokentype$$.types.colon) {
    let parent = this.curContext()
    if (parent === types.b_stat || parent === types.b_expr)
      return !parent.isExpr
  }
  if (prevType === $$tokentype$$.types._return)
    return $$whitespace$$.lineBreak.test(this.input.slice(this.lastTokEnd, this.start))
  if (prevType === $$tokentype$$.types._else || prevType === $$tokentype$$.types.semi || prevType === $$tokentype$$.types.eof || prevType === $$tokentype$$.types.parenR)
    return true
  if (prevType == $$tokentype$$.types.braceL)
    return this.curContext() === types.b_stat
  return !this.exprAllowed
}

pp.updateContext = function(prevType) {
  let update, type = this.type
  if (type.keyword && prevType == $$tokentype$$.types.dot)
    this.exprAllowed = false
  else if (update = type.updateContext)
    update.call(this, prevType)
  else
    this.exprAllowed = type.beforeExpr
}

// Token-specific context update code

$$tokentype$$.types.parenR.updateContext = $$tokentype$$.types.braceR.updateContext = function() {
  if (this.context.length == 1) {
    this.exprAllowed = true
    return
  }
  let out = this.context.pop()
  if (out === types.b_stat && this.curContext() === types.f_expr) {
    this.context.pop()
    this.exprAllowed = false
  } else if (out === types.b_tmpl) {
    this.exprAllowed = true
  } else {
    this.exprAllowed = !out.isExpr
  }
}

$$tokentype$$.types.braceL.updateContext = function(prevType) {
  this.context.push(this.braceIsBlock(prevType) ? types.b_stat : types.b_expr)
  this.exprAllowed = true
}

$$tokentype$$.types.dollarBraceL.updateContext = function() {
  this.context.push(types.b_tmpl)
  this.exprAllowed = true
}

$$tokentype$$.types.parenL.updateContext = function(prevType) {
  let statementParens = prevType === $$tokentype$$.types._if || prevType === $$tokentype$$.types._for || prevType === $$tokentype$$.types._with || prevType === $$tokentype$$.types._while
  this.context.push(statementParens ? types.p_stat : types.p_expr)
  this.exprAllowed = true
}

$$tokentype$$.types.incDec.updateContext = function() {
  // tokExprAllowed stays unchanged
}

$$tokentype$$.types._function.updateContext = function(prevType) {
  if (prevType.beforeExpr && prevType !== $$tokentype$$.types.semi && prevType !== $$tokentype$$.types._else &&
      !((prevType === $$tokentype$$.types.colon || prevType === $$tokentype$$.types.braceL) && this.curContext() === types.b_stat))
    this.context.push(types.f_expr)
  this.exprAllowed = false
}

$$tokentype$$.types.backQuote.updateContext = function() {
  if (this.curContext() === types.q_tmpl)
    this.context.pop()
  else
    this.context.push(types.q_tmpl)
  this.exprAllowed = false
}
exports.TokContext = TokContext, exports.types = types;